/*******************************************************************************
 * Copyright (c) 2010 Attila Kovacs <attila_kovacs@post.harvard.edu>.
 * All rights reserved. 
 * 
 * This file is part of minicrush.
 * 
 *     minicrush is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 * 
 *     minicrush is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with minicrush.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *     Attila Kovacs <attila_kovacs@post.harvard.edu> - initial API and implementation
 ******************************************************************************/
import java.io.*;
import java.util.*;

public class DataSimulator {
	
	static Hashtable<String, String> options;
	
	public static void main(String[] args) {	
		Locale.setDefault(Locale.US);
	
		String[] initargs = new String[args.length + 3];
		initargs[0] = args[0];
		initargs[1] = "-forget=aclip";
		initargs[2] = "-forget=vclip";
		initargs[3] = "-forget=downsample";
		System.arraycopy(args, 1, initargs, 4, args.length-1);
		
		MiniCRUSH crush = MiniCRUSH.init(initargs);
		options = crush.options;
		SkyMap reduced = new SkyMap();
		try { reduced.projection = SphericalProjection.forName(options.get("projection")); }
		catch(Exception e) { e.printStackTrace(); }
		
		crush.map = reduced;
			
		SphericalProjection projection = new RadioProjection();
		projection.setReference(crush.subscan[0][0].equatorial);
		reduced.projection = projection;
		
		for(int i=0; i<crush.scans; i++) {
			for(int k=0; k<crush.subscan[i].length; k++) {
				LabocaScan scan = crush.subscan[i][k];

				if(options.containsKey("jackknife")) {
					System.err.println("Jackknifing...");
					if(Math.random() < 0.5) scan.scale(-1.0);
				}
				
				scan.map = reduced;
					
				if(!options.containsKey("orig")) {
					System.err.println("Clearing Scan Data.");
					scan.clearData();
					
					if(options.containsKey("uniform")) {
						System.err.println("Using uniform gains.");
						for(int c=0; c<scan.channels; c++) {
							scan.channel[c].gain = 1.0;
							scan.channel[c].mainBeamEfficiency = 1.0;
						}
					}

					scan.map = reduced;

					if(!(options.containsKey("noiseless") || options.containsKey("orig"))) {
						System.err.println("Adding White Noise.");
						scan.randomData();
					}

					if(options.containsKey("constant")) {
						System.err.println("Adding Constant value to all channels.");
						double value = Double.parseDouble(options.get("constant"));
						scan.offset(value * scan.instrumentGain);
					}
					if(options.containsKey("decorrelate")) addCorrelated(scan);
					if(options.containsKey("wafers")) addWafers(scan);
					if(options.containsKey("boxes")) addBoxes(scan);
					if(options.containsKey("cables")) addCables(scan);
					if(options.containsKey("amps")) addAmps(scan);
				}
				
				if(options.containsKey("sources")) {
					try { addSources(scan, options.get("sources")); }
					catch(IOException e) { System.err.println("ERROR! Reading sources file..."); }
				}
				
				try {
					scan.writeData(options.containsKey("overwrite") ? "" : ".sim");
				}
				catch(Exception e) { 
					System.err.println("Problem writing data."); 
					e.printStackTrace();
				}
			}
		}
		
	}
	
	
	public static void addSources(LabocaScan scan, String fileName) throws IOException {
		Vector<GaussianSource> sources = new Vector<GaussianSource>();
		BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(fileName)));
		String line = null;
		while((line = in.readLine()) != null) if(line.length() > 0) if(line.charAt(0) != '#') {
			try { sources.add(new GaussianSource(line)); }
			catch(IllegalArgumentException e) { e.printStackTrace(); }			
		}
		System.err.println("Adding " + sources.size() + " Gaussian source(s).");
		scan.addSources(sources);
	}
	
	public static void addCorrelated(LabocaScan scan) {
		System.err.println("Adding Correlated Noise to " + scan.allChannels.length + " channels.");
			float[] gain = new float[scan.allChannels.length];
		for(int c=0; c<gain.length; c++) gain[c] = (float)scan.channel[c].gain;
		scan.addCorrelated(scan.allChannels, gain, getLevel(scan, "decorrelate", 3e-5));
	}
	
	public static double getLevel(LabocaScan scan, String option, double defaultValue) {
		String value = options.get(option);
		return (value.length() > 0 ?Double.parseDouble(value) : defaultValue) * Math.sqrt(Unit.sec / scan.integrationTime);
	}
	
	public static void addCables(LabocaScan scan) {
		System.err.println("Adding Correlated Cables.");	
		addCorrelated(scan, scan.cableChannels, getLevel(scan, "cables", 5e-6));
	}
	
	public static void addWafers(LabocaScan scan) {
		System.err.println("Adding Correlated Wafers.");	
		addCorrelated(scan, scan.waferChannels, getLevel(scan, "wafers", 2e-6));
	}
	
	public static void addBoxes(LabocaScan scan) {
		System.err.println("Adding Correlated Boxes.");
		addCorrelated(scan, scan.boxChannels, getLevel(scan, "boxes", 2e-6));
	}
	
	public static void addAmps(LabocaScan scan) {
		System.err.println("Adding Correlated Amps.");
		addCorrelated(scan, scan.boxChannels, getLevel(scan, "amps", 1e-6));
	}
	
	
	public static void addCorrelated(LabocaScan scan, int[][] channelSelection, double nyquistLevel) {
		for(int g=0; g<channelSelection.length; g++) {
			float[] gain = new float[channelSelection[g].length];
			Arrays.fill(gain, 1.0F);
			scan.addCorrelated(channelSelection[g], gain, nyquistLevel);			
		}
	}	
}
