/*******************************************************************************
 * Copyright (c) 2010 Attila Kovacs <attila_kovacs@post.harvard.edu>.
 * All rights reserved. 
 * 
 * This file is part of minicrush.
 * 
 *     minicrush is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 * 
 *     minicrush is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with minicrush.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *     Attila Kovacs <attila_kovacs@post.harvard.edu> - initial API and implementation
 ******************************************************************************/
// Copyright (c) 2007 Attila Kovacs 

import java.io.*;
import java.util.*;

public abstract class Interpolator extends Vector<InterpolatorData> {
	/**
	 * 
	 */
	private static final long serialVersionUID = -7962217110619389946L;

	public Interpolator(String fileName) throws IOException {
		readData(fileName);
		System.err.println(getClass().getSimpleName() + "> " + size() + " records parsed.");	
		Collections.sort(this);
	}
	
	public abstract void readData(String fileName) throws IOException; 
	
	// Linear interpolation.
	// Throws Exception if MJD is outside of the interpolator range.
	public double getValue(double MJD) throws ArrayIndexOutOfBoundsException {
		int upper = getIndexAbove(MJD);
		
		double dt1 = MJD - get(upper-1).MJD;
		double dt2 = get(upper).MJD - MJD;
		
		return (dt2 * get(upper-1).value + dt1 * get(upper).value) / (dt1 + dt2);	
	}	
	
	public int getIndexAbove(double MJD) throws ArrayIndexOutOfBoundsException {
		int lower = 0, upper = size()-1;
		
		if(MJD < get(lower).MJD || MJD > get(upper).MJD) 
			throw new ArrayIndexOutOfBoundsException(getClass().getSimpleName() + "> outside of interpolator range.");
		
		while(upper - lower > 1) {
			int i = (upper + lower) / 2;
			double iMJD = get(i).MJD;
			if(MJD >= iMJD) lower = i;
			if(MJD <= iMJD) upper = i;
		}
		
		return upper;
	}
	
	public double getSmoothValue(double MJD, double deltat) throws ArrayIndexOutOfBoundsException {
		int i0 = getIndexAbove(MJD); 
		
		double sum = 0.0, sumw = 0.0;
		InterpolatorData last = get(i0);
		
		deltat /= Unit.day;
		double sigmat = deltat / 2.35;
		double A = -0.5 / (sigmat * sigmat);
		double dt;
		
		int i = i0;
		while(i < size() && (dt = last.MJD - MJD) < 2 * deltat) {
			double w = Math.exp(-A*dt*dt);
			sum += w * last.value;
			sumw += w;
			last = get(++i);
		}
		
		i = i0-1;
		last = get(i0);
		while(i >= 0 && (dt = MJD - last.MJD) < 2 * deltat) {
			double w = Math.exp(A*dt*dt);
			sum += w * last.value;
			sumw += w;
			last = get(--i);
		}
		
		return sum / sumw;
	}
}

