/*******************************************************************************
 * Copyright (c) 2010 Attila Kovacs <attila_kovacs@post.harvard.edu>.
 * All rights reserved. 
 * 
 * This file is part of minicrush.
 * 
 *     minicrush is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 * 
 *     minicrush is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with minicrush.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Contributors:
 *     Attila Kovacs <attila_kovacs@post.harvard.edu> - initial API and implementation
 ******************************************************************************/
// Copyright (c) 2007 Attila Kovacs 

public class LabocaFrame implements Cloneable {
	LabocaScan scan;
	EquatorialCoordinates equatorial;
	HorizontalCoordinates horizontal;
	Vector2D horizontalOffset;
	
	// TODO check on PA sign convention...
	double MJD, LST, PA, cosPA, sinPA, chopperPosition;
	double cosA = 1.0, sinA = 0.0; 
	double he3Temp;
	int chopperPhase;
	
	double zenithTau = 0.0;
	double transmission = 1.0;
	
	int flag = 0;
	double relativeWeight = 1.0;
	double dependents = 0.0;
	double lastDependents = 0.0;
	double temp;
	double dof = 1.0;
	
	float[] data;
	byte[] sampleFlag;
	
	public LabocaFrame(LabocaScan parent) { scan = parent; }
	
	static double skydipFactor = 1.0;
	
	public Object clone() {
		try { return super.clone(); }
		catch(CloneNotSupportedException e) { return null; }
	}
	
	public void slim(LabocaPixel[] newPixels) {
		float[] reduced = new float[newPixels.length];
		byte[] newSampleFlag = new byte[newPixels.length];

		for(int k=0; k<newPixels.length; k++) {
			final int oldk = newPixels[k].index;
			reduced[k] = data[oldk];
			newSampleFlag[k] = sampleFlag[oldk];
		}
			
		data = reduced;
		sampleFlag = newSampleFlag;
	}

	public void calcParallacticAngle() {
		setPA(horizontal.getParallacticAngle(scan.site));
	}
	
	public void setPA(double angle) {
		PA = angle;
		sinPA = Math.sin(PA);
		cosPA = Math.cos(PA);
	}
	
	public void setAngle(double angle) {
		sinA = Math.sin(angle);
		cosA = Math.cos(angle);
	}
	
	public final double getX(final LabocaPixel pixel) {
		return cosA * pixel.position.x - sinA * pixel.position.y;	
	}
	
	public final double getY(final LabocaPixel pixel) {
		return cosA * pixel.position.y + sinA * pixel.position.x;	
	}	
	
	public EquatorialCoordinates getApparentEquatorial() {
		EquatorialCoordinates apparent = equatorial.copy();
		apparent.precess(JulianEpoch.getYearForMJD(MJD));
		return apparent;
	}
	
	public void calcHorizontal() {
		horizontal = getApparentEquatorial().toHorizontal(scan.site, LST);
	}
	
	public void calcEquatorial() {
		// This assumes that the object is tracked on sky...
		if(scan.isTracking) {
			if(equatorial == null) equatorial = (EquatorialCoordinates) scan.equatorial.clone();
			equatorial.setNativeLongitude(scan.equatorial.x + (cosPA * horizontalOffset.x - sinPA * horizontalOffset.y) / scan.equatorial.cosLat);
			equatorial.setNativeLatitude(scan.equatorial.y + (cosPA * horizontalOffset.y + sinPA * horizontalOffset.x));	
		}
		else {
			equatorial = horizontal.toEquatorial(scan.site, LST);
			equatorial.epoch.setMJD(MJD);
			equatorial.precess(scan.equatorial.epoch);
		}
	}

	public void validate() {
		if(sampleFlag == null) sampleFlag = new byte[data.length];
		else if(sampleFlag.length != data.length) sampleFlag = new byte[data.length];
		
		if(LST < -900.0*Unit.deg) flag |= SKIP_SOURCE;
		else if(horizontalOffset.x < -900.0*Unit.deg) flag |= SKIP_SOURCE;
		else if(horizontalOffset.y < -900.0*Unit.deg) flag |= SKIP_SOURCE;
	}
	
	public void setZenithTau(double value) {
		zenithTau = value;
		transmission = Math.exp(-skydipFactor * zenithTau/horizontal.sinLat);		
	}

	public final void getEquatorial(final LabocaPixel pixel, final EquatorialCoordinates coords) {
		final double x = getX(pixel);
		final double y = getY(pixel);
		
		// The proper GLS convention uses actual cos(DEC)
		// However, APECS uses cos(DEC0)
		coords.setNativeLongitude(equatorial.x + (cosPA * x - sinPA * y) / scan.equatorial.cosLat);
		coords.setNativeLatitude(equatorial.y + (cosPA * y + sinPA * x));
	}
	
	
	public final void getHorizontal(final LabocaPixel pixel, final HorizontalCoordinates coords) {
		// The proper GLS convention uses actual cos(DEC)
		// However, APECS uses cos(DEC0)
		coords.setNativeLongitude(horizontal.x + getX(pixel) / scan.horizontal.cosLat);
		coords.setNativeLatitude(horizontal.y + getY(pixel));
	}
	
	
	public final void getHorizontalOffset(final LabocaPixel pixel, final Vector2D offset) {
		offset.x = horizontalOffset.x + getX(pixel);
		offset.y = horizontalOffset.y + getY(pixel);
	}
	
	
	public final void getEquatorialOffset(final LabocaPixel pixel, final Vector2D offset) {
		getHorizontalOffset(pixel, offset);
		toEquatorial(offset);
	}
	
	public final void getEquatorialOffset(final Vector2D offset) {
		equatorial.getOffsetFrom(scan.equatorial, offset);
	}
	
	// Rotate by PA
	public final void toEquatorial(Vector2D offset) {
		final double x = offset.x;
		offset.x = cosPA * x - sinPA * offset.y;
		offset.y = cosPA * offset.y + sinPA * x;
		offset.x *= -1.0;
	}
	
	// Rotate by -PA
	public final void toHorizontal(Vector2D offset) {
		final double x = -offset.x;
		offset.x = cosPA * x + sinPA * offset.y;
		offset.y = cosPA * offset.y - sinPA * x;
	}
	
	
	public static byte SAMPLE_SOURCE_BLANK = 1;
	public static byte SAMPLE_SPIKE = 1<<1;
	public static byte SAMPLE_SPIKY_NEIGHBOUR = 1<<2;
	public static byte SAMPLE_SPIKY_FEATURE = 1<<3;
	public static byte SAMPLE_SPIKE_FLAGS = (byte) (SAMPLE_SPIKE | SAMPLE_SPIKY_NEIGHBOUR | SAMPLE_SPIKY_FEATURE);
	public static byte SAMPLE_SKIP = 1<<4;
	public static byte SAMPLE_PHOTOMETRY = 1<<5;
	
	public static int FLAG_DISCARD = 1;
	public static int FLAG_SPIKY = 1<<1;
	public static int FLAG_DOF = 1<<2;
	public static int FLAG_WEIGHT = 1<<3;

	public static int SKIP_SOURCE = 1<<8;
	public static int SKIP_MODELS = 1<<9;
	public static int SKIP_WEIGHTING = 1<<10;
	//public static int SKIP_SYNCHING = 1<<11;
	
	
	public static int CHOP_LEFT = 1<<16;
	public static int CHOP_RIGHT = 1<<17;
	public static int CHOP_TRANSIT = 1<<18;
	public static int CHOP_FLAGS = CHOP_LEFT | CHOP_RIGHT | CHOP_TRANSIT;
	

	public static int BAD_DATA = FLAG_SPIKY | FLAG_DISCARD;
	public static int MODELING_FLAGS = SKIP_MODELS | BAD_DATA | FLAG_DOF | FLAG_WEIGHT;
	public static int SOURCE_FLAGS = SKIP_SOURCE | MODELING_FLAGS;
	public static int WEIGHTING_FLAGS = SKIP_WEIGHTING | MODELING_FLAGS;
	
	
}
